/*
*****************************************************************************
  *		XpressNetMaster.h - library for XpressNet protocoll
  *		Copyright (c) 03/2016 Philipp Gahtow  All right reserved.
  *
  *		Version 1.7
*****************************************************************************
  * FUNKTIONS:
  *
  *		Build a XpressNet Master for driving and switch
  *		Support for Roco LokMaus2 and Multimaus
  *		Basic support for direct CV programming
  *		Check XOR before packet dekoding
  * 
  * 	-> Please contact Lenz Inc. for more details about XpressNet.
  *
  * not yet Supported:
  *	- Emergency stop a locomotive: 0x92 AddrH AddrL [XOR]
  *	- Emergency stop a locomotive: 0x91 loco_addr [XOR] (v2)
  *	- Service Mode: 0x21, 0x22, 0x23, 0x24, 0xE6
  *	- DCC extended accessory command: 0x13 0x01 B+AddrH AddrL
  *	- DCC FAST CLOCK: 0x01 0xF2 0xF3
  *	- Command Tunnel: 0x3* ** ** [XOR]
  *	- BiDi messages: 0x7* ** ** [XOR]
  *	- Library Entry: 0xE9 0xF1 AddrH AddrL IDX SIZE [NAME][XOR]
  *	- Set function state: 0xE4 0x2* AddrH AddrL Group [XOR]
*****************************************************************************
*/

// include this library's description file
#include "XpressNetMaster.h"
#include <avr/interrupt.h>

//XpressNetMasterClass *XpressNetMasterClass::active_object = 0;	//Static

// Constructor /////////////////////////////////////////////////////////////////
// Function that handles the creation and setup of instances

XpressNetMasterClass::XpressNetMasterClass()
{
	// initialize this instance's variables 
	XNetAdr = 0;	//Startaddresse des ersten XNet Device
	XSendCount = micros();		//Reset Sendercount for next Call Byte
	XNetclear();	//alte Nachricht lschen
	Railpower = csTrackVoltageOff;
	getNextXNetAdr();
	Fahrstufe = Loco128;
}

//******************************************Serial*******************************************
void XpressNetMasterClass::setup(uint8_t FStufen, uint8_t  XControl)  //Initialisierung Serial
{
	Fahrstufe = FStufen;
	MAX485_CONTROL = XControl;
	// LISTEN_MODE 
	pinMode(MAX485_CONTROL, OUTPUT);
	digitalWrite (MAX485_CONTROL, LOW);

	//Set up on 62500 Baud
	cli();  //disable interrupts while initializing the USART
	#ifdef __AVR_ATmega8__
	 UBRRH = 0;
	 UBRRL = 0x0F;
	 UCSRA = 0;
	 UCSRB = (1<<RXEN) | (1<<TXEN) | (1<<RXCIE) | (1<<UCSZ2);
	 UCSRC = (1<<UCSZ1) | (1<<UCSZ0);
	#else
	#ifdef SERIAL_PORT_0
	 UBRR0H = 0;
	 UBRR0L = 0x0F;
	 UCSR0A = 0;
	 UCSR0B = (1<<RXEN0) | (1<<TXEN0) | (1<<RXCIE0) | (1<<UCSZ02);
	 UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);
	 #else
	 UBRR1H = 0;
	 UBRR1L = 0x0F;
	 UCSR1A = 0;
	 UCSR1B = (1<<RXEN1) | (1<<TXEN1) | (1<<RXCIE1) | (1<<UCSZ12);
	 UCSR1C = (1<<UCSZ11) | (1<<UCSZ10);
	 #endif
	#endif
	sei(); // Enable the Global Interrupt Enable flag so that interrupts can be processed 
	 /*
	 *  Enable reception (RXEN = 1).
	 *  Enable transmission (TXEN0 = 1). 
	 *	Enable Receive Interrupt (RXCIE = 1).
	 *  Set 8-bit character mode (UCSZ00, UCSZ01, and UCSZ02 together control this, 
	 *  But UCSZ00, UCSZ01 are in Register UCSR0C).
	 */
}

// Public Methods //////////////////////////////////////////////////////////////
// Functions available in Wiring sketches, this library, and other libraries

//*******************************************************************************************
//Daten ermitteln und Auswerten
void XpressNetMasterClass::update(void) 
{
	if (micros() - XSendCount >= XNetTimeUntilNext) {

		if (XNetReceive() == true && rxXOR == XNetMsg[XNetMsg[XNetlength]]) {	//Data, vorhanden und XOR korrekt!
			/*
			Serial.print("RX: ");
			Serial.print(XNetMsg[XNetheader], HEX);
			Serial.print("-");
			Serial.print(XNetMsg[XNetdata1], HEX);
			Serial.print(" ");
			Serial.print(XNetMsg[XNetdata2], HEX);
			Serial.print(" ");
			Serial.print(XNetMsg[XNetdata3], HEX);
			Serial.print(" ");
			Serial.print(XNetMsg[XNetdata4], HEX);
			Serial.print("..");
			*/
			switch (XNetMsg[XNetheader]) {
			case 0x21:
				if (XNetMsg[XNetdata1] == 0x24 && XNetMsg[XNetdata2] == 0x05) {		//Command station status indication response
					/*
					Bit 0: =1 - Command station is in emergency off
					Bit 1: =1 - Command station is in emergency stop
					Bit 2: Command station-Start mode (0 = manual mode, 1 = automatic mode)
					Automatic Mode: All locomotives start at their last known speed each
					time the command station is powered up
					Manual Mode: All locomotives have a speed of 0 and functions out on
					command station power up
					Bit 3: = 1 - The command station is in service mode
					Bit 4: reserved
					Bit 5: reserved
					Bit 6: = 1 - The command station is performing a power up.
					Bit 7: = 1 - There was a RAM check error in the command station
					*/
					byte status = 0x01;	//csTrackVoltageOff = B1;
					switch (Railpower) {
						case csNormal:			status = 0; break;
						case csEmergencyStop:	status = B10; break;
						case csServiceMode:		status = B1000; break;
					}
					unsigned int sendStatus[] = { DirectedOps, 0x62, 0x22, status, 0x00 };
					getXOR(sendStatus, 5);
					XNetsend(sendStatus, 5);
/*
					Serial.print("TX: ");
					Serial.print(sendStatus[1], HEX);
					Serial.print("-");
					Serial.print(sendStatus[2], HEX);
					Serial.print(" ");
					Serial.print(sendStatus[3], HEX);
					Serial.print(" ");
					Serial.println(sendStatus[4], HEX);
*/
				}
				if (XNetMsg[XNetdata1] == 0x21 && XNetMsg[XNetdata2] == 0x00) {		//Command station softwareversion response
					unsigned int sendVersion[] = { DirectedOps, 0x63, 0x21, XNetVersion, XNetID, 0x00 }; //63-21 36 0 74
					getXOR(sendVersion, 6);
					XNetsend(sendVersion, 6);
/*
					Serial.print("TX: ");
					Serial.print(sendVersion[1], HEX);
					Serial.print("-");
					Serial.print(sendVersion[2], HEX);
					Serial.print(" ");
					Serial.print(sendVersion[3], HEX);
					Serial.print(" ");
					Serial.print(sendVersion[4], HEX);
					Serial.print(" ");
					Serial.println(sendVersion[5], HEX);
*/
				}
				if (XNetMsg[XNetdata1] == 0x80 && XNetMsg[XNetdata2] == 0xA1) {		//Alles Aus (Notaus)
					Railpower = csTrackVoltageOff;	//Save Power State
//					setPower(XNetTrackVoltageOff);
					if (notifyXNetPower)
						notifyXNetPower(csTrackVoltageOff);
				}
				if (XNetMsg[XNetdata1] == 0x81 && XNetMsg[XNetdata2] == 0xA0) {		//Alles An
					Railpower = csNormal;	//Save Power State
//					setPower(Railpower);
					if (notifyXNetPower)
						notifyXNetPower(csNormal);
				}
				XNetclear();	//alte Nachricht lschen
				return; // break;
			case 0x23:
				if (XNetMsg[XNetdata1] == 0x17) { //0x23 0x17 CV DAT[XOR] - Paged Mode write request(Paged mode)
//					setPower(csServiceMode);
					if (notifyXNetDirectCV)
						notifyXNetDirectCV(XNetMsg[XNetdata2], XNetMsg[XNetdata3]);
				}
				XNetclear();
				return; 
			case 0x80:
				if (XNetMsg[XNetdata1] == 0x80) {		//EmStop
					Railpower = csEmergencyStop;	//Save Power State
//					setPower(Railpower);
					if (notifyXNetPower)
						notifyXNetPower(csEmergencyStop);
					XNetclear();	//alte Nachricht lschen
					return; // break;
				}
			case 0xE3:
				if (XNetMsg[XNetdata1] == 0x00) {	//Lokinformationen & F0 bis F12 anfordern
					if (notifyXNetgiveLocoInfo)
						notifyXNetgiveLocoInfo(DirectedOps, word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]));
				}
				if (XNetMsg[XNetdata1] == 0x07) {	//Funktionststaus F13 bis F28 anfordern
					//0x07, sonst ist LokMaus2 langsam!
					if (notifyXNetgiveLocoFunc)
						notifyXNetgiveLocoFunc(DirectedOps, word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]));
				}
				XNetclear();	//alte Nachricht lschen
				return;
			case 0xE4:	//Fahrbefehle
				if (XNetMsg[XNetdata1] == 0x10) {	//14 Fahrstufen
					if (notifyXNetLocoDrive14)
						notifyXNetLocoDrive14(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x11) {	//27 Fahrstufen
					if (notifyXNetLocoDrive27)
						notifyXNetLocoDrive27(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x12) {	//28 Fahrstufen
					if (notifyXNetLocoDrive28)
						notifyXNetLocoDrive28(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x13) {	//128 Fahrstufen
					if (notifyXNetLocoDrive128)
						notifyXNetLocoDrive128(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x20) {	//Funktionsbefehl Gruppe1 0 0 0 F0 F4 F3 F2 F1
					if (notifyXNetLocoFunc1)
						notifyXNetLocoFunc1(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x21) {	//Funktionsbefehl Gruppe2 0000 F8 F7 F6 F5
					if (notifyXNetLocoFunc2)
						notifyXNetLocoFunc2(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x22) {	//Funktionsbefehl Gruppe3 0000 F12 F11 F10 F9
					if (notifyXNetLocoFunc3)
						notifyXNetLocoFunc3(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x23 || XNetMsg[XNetdata1] == 0xF3) {	//Funktionsbefehl Gruppe4 F20-F13
					//0xF3 = undocumented command is used when a mulitMAUS is controlling functions f20..f13. 
					if (notifyXNetLocoFunc4)
						notifyXNetLocoFunc4(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				else if (XNetMsg[XNetdata1] == 0x28) {	//Funktionsbefehl Gruppe5 F28-F21
					if (notifyXNetLocoFunc5)
						notifyXNetLocoFunc5(word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]), XNetMsg[XNetdata4]);
				}
				//Update permernent Multimaus Status!
				//SetLocoBusy(DirectedOps, word(XNetMsg[XNetdata2], XNetMsg[XNetdata3]));	//set Loco Busy, to make Multimaus request status!
				XNetclear();	//alte Nachricht lschen
				return;
			case 0x42:	//Accessory Decoder information request
				if (notifyXNetTrntInfo)
					notifyXNetTrntInfo(DirectedOps, XNetMsg[XNetdata1], XNetMsg[XNetdata2]);
				XNetclear();	//alte Nachricht lschen
				return;
			case 0x52:	//Accessory Decoder operation request
				if (notifyXNetTrnt)
					notifyXNetTrnt((XNetMsg[XNetdata1] << 2) | ((XNetMsg[XNetdata2] & B110) >> 1), XNetMsg[XNetdata2]);
					//XNetMsg[XNetdata2] = 0000 ABBP
					//A = Weichenausgang(Spulenspannung EIN/AUS)
					//BB = Adresse des Dekoderport 1..4
					//P = Ausgang (Gerade = 0 / Abzweigen = 1)
				XNetclear();	//alte Nachricht lschen
				return;
			default:  //Befehl in Zentrale nicht vorhanden
				if (XNetMsg[XNetheader] != 0x00) {
					unsigned int NotIn[] = { DirectedOps, 0x61, 0x82, 0x00 };	
					getXOR(NotIn, 4);
					XNetsend(NotIn, 4);
				}
			}
		}
		XNetclear();	//alte Nachricht lschen
		//Send CallByteInquiry for next Addr:
		getNextXNetAdr();
		unsigned int NormalInquiry[] = { CallByteInquiry };
		XNetsend( NormalInquiry, 1);
	}
}

//--------------------------------------------------------------------------------------------
void XpressNetMasterClass::getNextXNetAdr(void)
{
	XNetAdr++;		//nchste Adresse im XNet
	if (XNetAdr > 31)	//wenn letzte erreicht von Beginn!
		XNetAdr = 1;

	/*
	Call Byte Definitions:
	Call_Response 					P11A AAAA
	Call_Inquiry					P10A AAAA
	Call_Request_Ack_From_Device	P00A AAAA
	Call_Broadcast					P110 0000
	Call_Transfer_Err				P11A AAAA
	*/
	CallByteInquiry = callByteParity(XNetAdr | 0x40) | 0x100;	// the address for Call Byte Window
	RequestAck = callByteParity(XNetAdr | 0x00) | 0x100;		// the address for a request acknowlegement sent
	DirectedOps = callByteParity(XNetAdr | 0x60) | 0x100;		// the address when we are sending ops
}

//--------------------------------------------------------------------------------------------
//Zustand der Gleisversorgung setzten
void XpressNetMasterClass::setPower(byte Power) 
{
  Railpower = Power;	//Save Power State
  switch (Power) {	
  case csNormal: {
		unsigned int NormalPower[] = { GENERAL_BROADCAST, 0x61, 0x01, 0x60 };
		XNetsend(NormalPower, 4);
		return;
	}
  case csEmergencyStop: {
		unsigned int EStopPower[] = { GENERAL_BROADCAST, 0x81, 0x00, 0x81 };
		XNetsend(EStopPower, 4);
		return;
	}
  case csTrackVoltageOff: {
		unsigned int OffPower[] = { GENERAL_BROADCAST, 0x61, 0x00, 0x61 };
		XNetsend(OffPower, 4);
		return;
	}
  case csShortCircuit: {
		unsigned int ShortPower[] = { GENERAL_BROADCAST, 0x61, 0x08, 0x67 };
		XNetsend(ShortPower, 4);
		return;
	}
  case csServiceMode: {
		unsigned int ServicePower[] = { GENERAL_BROADCAST, 0x61, 0x01, 0x63 };
		XNetsend(ServicePower, 4);
		return;
	}
  }
  return;
}

//--------------------------------------------------------------------------------------------
//Lok in use (Busy)
void XpressNetMasterClass::SetLocoBusy(uint8_t UserOps, uint16_t Adr) {
	unsigned int LocoInfo[] = { UserOps | 0x100, 0xE3, 0x40, Adr >> 8, Adr & 0xFF, 0x00 };
	getXOR(LocoInfo, 6);
	XNetsend(LocoInfo, 6);
}

//--------------------------------------------------------------------------------------------
//Lokinfo an XNet Melden
void XpressNetMasterClass::SetLocoInfo(uint8_t UserOps, uint8_t Speed, uint8_t F0, uint8_t F1) {
	SetLocoInfo(UserOps, Fahrstufe, Speed, F0, F1);
}

void XpressNetMasterClass::SetLocoInfo(uint8_t UserOps, uint8_t Steps, uint8_t Speed, uint8_t F0, uint8_t F1) {
	//0xE4	| 0000 BFFF | RVVV VVVV | 000F FFFF | FFFF FFFF | XOr
	//LokInfo Identification Speed FA FB
	unsigned int LocoInfo[] = { UserOps | 0x100, 0xE4, B00001000 | Steps, Speed, F0, F1, 0x00 };
	getXOR(LocoInfo, 7);
	XNetsend(LocoInfo, 7);
}

//--------------------------------------------------------------------------------------------
//LokFkt an XNet Melden
void XpressNetMasterClass::SetFktStatus(uint8_t UserOps, uint8_t F4, uint8_t F5) {
	//F4 = F20-F13
	//F5 = F28-F21
	unsigned int LocoFkt[] = { UserOps | 0x100, 0xE3, 0x52, F4, F5, 0x00 };
	getXOR(LocoFkt, 6);
	XNetsend(LocoFkt, 6);
}

//--------------------------------------------------------------------------------------------
//TrntPos an XNet Melden
void XpressNetMasterClass::SetTrntStatus(uint8_t UserOps, uint8_t Address, uint8_t Data) {
	// data=ITTN ZZZZ	Z=Weichenausgang1+2 (Aktive/Inaktive);
	//0x42 AAAA AAAA ITTN ZZZZ
	//I = is 1, the switching command requested has not been completed
	//TT = turnout group (0-3)
	//N = N=0 is the lower nibble, N=1 the upper nibble
	//Z3 Z2 Z1 Z0 = (Z1,Z0|Z3,Z2) 01 "left", 10 "right"
	unsigned int TrntInfo[] = { UserOps | 0x100, 0x42, Address, Data, 0x00 };
	getXOR(TrntInfo, 5);
	XNetsend(TrntInfo, 5);
}

//--------------------------------------------------------------------------------------------
// send along a bunch of bytes to the Command Station
void XpressNetMasterClass::XNetsend(unsigned int *dataString, byte byteCount) {
	digitalWrite(MAX485_CONTROL, HIGH); 	//SEND_MODE
	for (byte i = 0; i < byteCount; i++) {
		XNetTransmit9(*dataString);	//Data Byte
		dataString++;
	}
	WAIT_FOR_XMIT_COMPLETE;
	digitalWrite(MAX485_CONTROL, LOW);		//LISTEN_MODE
	XSendCount = micros();		//Reset Sendercount for next Call Byte
}

//--------------------------------------------------------------------------------------------
// calculate the XOR
void XpressNetMasterClass::getXOR (unsigned int *data, byte length) {
	byte XOR = 0x00;
	data++;		//without CallByte
	for (byte i = 0; i < (length-2); i++) { 
        XOR = XOR ^ *data;
		data++;
    }
	*data = XOR;
}

//--------------------------------------------------------------------------------------------
// calculate the parity bit in the call byte for this guy
unsigned int XpressNetMasterClass::callByteParity (unsigned int me) {
 int parity = (1==0);
 unsigned int vv;
 me = me & 0x7f;
 vv = me;

 while (vv) {
       parity = !parity;
       vv = vv & (vv-1);
 }
 if (parity) me = me | 0x80;
 return me;
}

/*
//--------------------------------------------------------------------------------------------
void XpressNetMasterClass::XNetTransmit(unsigned char data8) {
 // wait for empty transmit buffer
 #ifdef __AVR_ATmega8__
  while (!(UCSRA & (1<<UDRE))) {}
 // put the data into buffer, and send 
 UDR = data8;
 #else
 #ifdef SERIAL_PORT_0
  while (!(UCSR0A & (1<<UDRE0))) {}
 // put the data into buffer, and send 
 UDR0 = data8;
 #else
 while (!(UCSR1A & (1<<UDRE1))) {}
 // put the data into buffer, and send
 UDR1 = data8;
 #endif
 #endif
}
*/

void XpressNetMasterClass::XNetTransmit9(unsigned int data8)
{
		/* wait for empty transmit buffer */
#ifdef __AVR_ATmega8__
		while (!(UCSRA & (1 << UDRE))) {}
		/* put the data into buffer, and send */
		UCSRB &= ~(1 << TXB8);
		if (data8 & 0x100)
			UCSRB |= (1 << TXB8);
		UDR = data8;
#else
#ifdef SERIAL_PORT_0
		while (!(UCSR0A & (1 << UDRE0))) {}
		/* put the data into buffer, and send */
		UCSR0B &= ~(1 << TXB80);
		if (data8 & 0x100)
			UCSR0B |= (1 << TXB80);
		UDR0 = data8;
#else
		while (!(UCSR1A & (1 << UDRE1))) {}
		/* put the data into buffer, and send */
		UCSR1B &= ~(1 << TXB81);
		if (data8 & 0x100) 
			UCSR1B |= (1 << TXB81);
		UDR1 = data8;
#endif
#endif
}

//--------------------------------------------------------------------------------------------
//Serial einlesen:
bool XpressNetMasterClass::XNetReceive(void)
{
	bool ok = false;
	// if there's any serial available, read it:
#ifdef __AVR_ATmega328P__
	while (Serial.available() > 0) {
		byte rxdata = Serial.read();
#else
#ifdef SERIAL_PORT_0
	while (Serial.available() > 0) {
		byte rxdata = Serial.read();
#else
	while (Serial1.available() > 0) {
		byte rxdata = Serial1.read();
#endif
#endif

		XSendCount = micros() + XNetTimeUntilNext;		//Reset Sendercount for next Call Byte
		if (XNetMsg[XNetlength] > 0)
			rxXOR = rxXOR ^ XNetMsg[XNetMsg[XNetlength]];		//XOR bestimmen
		XNetMsg[XNetlength]++;	    //weitere Nachrichtendaten
		XNetMsg[XNetMsg[XNetlength]] = rxdata;  //Eintragen
		
		ok = true;
	}
	return ok;
}

//--------------------------------------------------------------------------------------------
//Lschen des letzten gesendeten Befehls
void XpressNetMasterClass::XNetclear(void)
{
	XNetMsg[XNetlength] = 0x00;	//Reset Message
	XNetMsg[XNetheader] = 0x00;
	XNetMsg[XNetdata1] = 0x00;
	XNetMsg[XNetdata2] = 0x00;
	XNetMsg[XNetdata3] = 0x00;
	XNetMsg[XNetdata4] = 0x00;
	XNetMsg[XNetdata5] = 0x00;
	XNetMsg[XNetdata6] = 0x00;
	XNetMsg[XNetdata7] = 0x00;
	rxXOR = 0x00;	//Reset XOR
}
